# -*- coding: utf-8 -*-
"""
@author: julien de Rosny (c) Institut Langevin/ESPCI/CNRS
"""

from math import *
from numpy import *
from numpy.linalg import norm
from matplotlib.pyplot import *
from scipy.special import ellipeinc
from scipy.optimize import root
def MNonPara(a,b,k0=0):
    """
         Non parallel mutual inductance between wires
         a : [[x1,y1,z1],[x2,y2,z2]]
         b : [[x1,y1,z1],[x2,y2,z2]]
         
    """
    if (k0>0):
        return(MInte(a,b,k0))
    epsilon = 1e-7
    a = array(a)
    b = array(b)
    d11 = norm(a[0,:]-b[0,:])
    d12 = norm(a[0,:]-b[1,:]) #a1-b2)
    d21 = norm(a[1,:]-b[0,:]) #a2-b1)
    d22 = norm(a[1,:]-b[1,:]) #a2-b2)
    
    val =argmin([d11,d22,d12,d21])
    sign = 1
    if val == 1 or val == 3:
        a=a[::-1,:]
        sign = -1 #a1,a2 = a2,a1
    if val == 1 or val == 2:
        b=b[::-1,:] #b1,b2 = b2,b1
        sign = -1
    
    #arrow(a1[0],a1[1],a2[0]-a1[0],a2[1]-a1[1],width=.1)
    #arrow(b1[0],b1[1],b2[0]-b1[0],b2[1]-b1[1],width=.1)
    ll = norm(a[1,:]-a[0,:])#a2-a1)
    mm = norm(b[1,:]-b[0,:])#b2-b1)
    R1 = norm(b[1,:]-a[1,:])#b2-a2)
    R2 = norm(a[1,:]-b[0,:])#a2-b1) #ll
    R3 = norm(a[0,:]-b[0,:])#a1-b1) #0
    R4 = norm(b[1,:]-a[0,:])#b2-a1) # mm
    
    alpha2 = R4**2-R3**2+R2**2-R1**2
    
    try:
        if abs(R1-R3)>epsilon:
            cosepsilon = alpha2 / (2*ll*mm)
            mu = ll*(2*mm**2*(R2**2-R3**2-ll**2)+alpha2*(R4**2-R3**2-mm**2))/(4*ll**2*mm**2-alpha2**2)
            nu = mm*(2*ll**2*(R4**2-R3**2-mm**2)+alpha2*(R2**2-R3**2-ll**2))/(4*ll**2*mm**2-alpha2**2)
            M =  (mu+ll)  *  atanh(mm/(R1+R2)) \
               + (nu+mm)  *  atanh(ll/(R1+R4))
            if R3 > 0:
                M +=      -  mu      *  atanh(mm/(R3+R4))     \
                          -  nu      *  atanh(ll/(R2+R3))
            #else:
               # print('connected');
            M = 2 * cosepsilon * M * sign
        else:
            #print('parallel and identical wires')
            M =  Lnm(ll,R2) * sign
    except:
        print('a1, a2          : ',a1,a2)
        print('b1, b2          : ',b1,b2)
        print('alpha2,ll,mm    :',alpha2,ll,mm)
        print('R1,R2,R3,R4     :',R1,R2,R3,R4)
        print('mu,nu,cosepsilon:',mu,nu,cosepsilon)
        raise
        
    return(M)
    

    
#
def selfInducRect(l,w1,w2=0):
    """
        Self impedance of a wire of length l and rectangular cross section of sides w1 and w2
    """
    #return(2*l*(log(2*l/w)+0.5+0.2235*w/l)) 
    return(2*l*(log(2*l/(w1+w2))+0.5))
    
def selfInducTub(l,a_in,a_out):
    """
        Self impedance of a wire of length l and tubular cross-section of inner and outer radii equal to a_in, a_out
    """
    return(2*l*(log(2*l/a_out) + log(a_out/a_in) -1))
    
  
def Lnm(l,d,k0=0):
    """
    Mutual inductance of parallel wires of length l and d distant
    """
    if (k0==0):
        return(2*l*(log(l/d+sqrt(1+l**2/d**2))-sqrt(1+d**2/l**2)+d/l))
    
    r1 = array([[-l/2.,-d/2.,0.0],[+l/2.,-d/2.,0.0]])
    r2 = array([[-l/2., d/2.,0.0],[+l/2., d/2.,0.0]])
    return(MInte(r1,r2,k0))
    #return(2*l*(log(2*l/d)-1+d/l))
    
def I(theta,A,B):
    return(B**2*cos(theta)/(B**2*cos(theta)**2+A**2*sin(theta)**2))
    
def ellipse(theta,A,B):
    return([A*sin(theta),B*cos(theta)])
    
    

def G(r,r0,k0):
    #From EWA Ch15 p 719
    rnorm = np.linalg.norm(r-r0)
    G = np.exp(-1j*k0*rnorm)/(4*np.pi*rnorm)
    return(G)
    
    

def MInte(r1,r2,k0=1e-5):
    """
        Calculus of mutual inductance by integration.
        Warning : the inductance can be complex
        the result are expressed in 1e-7H
    """
    N1=10
    N2=10
    l1 = np.linalg.norm(r1[1,:]-r1[0,:])
    l2 = np.linalg.norm(r2[1,:]-r2[0,:])
    p1 = (r1[1,:]-r1[0,:])/l1 
    p2 = (r2[1,:]-r2[0,:])/l2
    d1 = l1/(N1-1)
    d2 = l2/(N2-1)
    champ2 = np.zeros((N2),dtype=complex)
    champ1 = np.zeros((N1),dtype=complex)

    for i1 in range(N1):
        r1c = r1[0,:] + p1*i1*d1 
        for i2 in range(N2):
            r2c = r2[0,:] + p2*i2*d2
            champ2[i2] = np.dot(p1,p2)*G(r1c,r2c,k0)
        champ1[i1] = np.trapz(champ2,dx=d2)   
    return( np.trapz(champ1,dx=d1)*4*np.pi   )


if __name__=='__main__':
    A = 12.6e-2
    B = 7.6e-2
    a1 = 0.123e-2 
    a2 = 0.198e-2
    lengthRod = 25e-2
    widthRing = 6e-2/4
    
    angles = zeros((12))
      
    uniform_distribution =  ellipeinc(2*pi,B/A)/len(angles)
    
    
    for i in range(len(angles)):
        angles[i]=root(lambda x:ellipeinc(x[0],B/A)-uniform_distribution*(i+1/2),[0]).x[0]
    
    if True:
        angles = array([ 75 , 48.2, 27.2, 9.2])
        angles = concatenate((180-angles[::-1],angles,-angles[::-1],angles[::]-180))
        angles = angles[::-1]/180.0*pi
    
    xy = np.zeros((2,len(angles)))
    xy = [ellipse(angle,A,B) for angle in angles]
    x = array([toto[0] for toto in xy ])
    y = array([toto[1] for toto in xy ])
    
    diffe = [sqrt((x[k]-x[k+1])**2+(y[k]-y[k+1])**2) for k in range(len(x)-1)]
    
    plot(x,y,'+')
    R0 = 33e-2/2
    R = sqrt(x**2+y**2)
    Rp = R0**2/R
    
    xp = Rp*sin(angles)
    yp = Rp*cos(angles)
    plot(xp,yp,'o')
    nb = 4
    L = zeros((nb))
    Lr = zeros((nb))
    
    IAB = I(angles,A,B)
    Iring = append(cumsum(IAB),0)
    
    for i in range(4):
        indice = 11-i
        I0 = IAB[indice]
        print('--')
        # Loop over the legs
        for k in range(len(angles)):
            d = sqrt((x[indice]-x[k])**2+(y[indice]-y[k])**2)
            dp = sqrt((x[indice]-xp[k])**2+(y[indice]-yp[k])**2)
            if d>0:
                L[i] += IAB[k]/I0*(Lnm(lengthRod,d)-Lnm(lengthRod,dp))
                print('%f %f %f'%(d,dp,(Lnm(lengthRod,d)-Lnm(lengthRod,dp))))
            else:
                L[i] += selfInducTub(lengthRod,a1,a2)-Lnm(lengthRod,dp)
        #arrow(x[indice],y[indice],x[(indice+1)%len(angles)]-x[indice],y[(indice+1)%len(angles)]-y[indice],width=.1)
        # Loop over the legs
        I0 = Iring[indice]
    
        for k in range(len(angles)):
            #arrow(x[k],y[k],x[(k+1)%len(angles)]-x[k],y[(k+1)%len(angles)]-y[k],width=.1)
            
            if k != indice:
                r1 = array([[x[indice],y[indice],0],[x[(indice+1)%len(angles)],y[(indice+1)%len(angles)],0]])
                r2 = array([[x[k],y[k],0],[x[(k+1)%len(angles)],y[(k+1)%len(angles)],0]])
                Lr[i] += abs(Iring[k]/I0*MNonPara(r1,r2))
            if k == i:
                l = norm(array([x[k],y[k]])-array([x[(k+1)%len(angles)],y[(k+1)%len(angles)]]))
                print('l=',l)
                Lr[i] += selfInducRect(l,widthRing)
    



    print("Test 2")
    
    L = 1.0
    d = 0.1
    k0 = 0.001
    l1 = array([[0.,0.,0.],[L,0.,0.]])
    l2=  array([[0.,0.,d],[L,0.,d]])
    M=MNonPara(l1,l2)
    print('Book : %e \nInte : %e\n'%(MNonPara(l1,l2),real(MInte(l1,l2,k0))))
     
    angle = 135.0*np.pi/180.0
    d = 1.0
    l = 1.0
    l1 = array([[d,0.0,0.],[d+l,0.,0.]])
    l2 = array([[d*cos(angle),d*sin(angle),0.],[(d+l)*cos(angle),(d+l)*sin(angle),0.]])
    print('Book : %e \nInte : %e\n'%(MNonPara(l1,l2),real(MInte(l1,l2,k0))))

    